#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include <stdio.h>
#include <stdlib.h>

#include "ST7735R_TFT.h"
#include "spi.h"

// ecranele din aplicatie
#define START_SCREEN 1
#define PLAY_SCREEN  2
#define END_SCREEN   3

// directiile sarpelui
#define NONE    0
#define LEFT    1
#define UP      2
#define DOWN    3
#define RIGHT   4

#define SNAKE_START_SIZE  7
#define SNAKE_MAX_SIZE    128

#define PIXEL_DIM         4
#define BLOCK_OFFSET      (PIXEL_DIM / 2)

#define BUFFER_SIZE       128

typedef struct {
    int16_t lin, col;

    uint8_t r, g, b;
} pixel_t;

pixel_t snake[SNAKE_MAX_SIZE];
uint16_t snake_len = SNAKE_START_SIZE;

pixel_t food;
int16_t regenerate_time;

uint8_t current_state, prev_state, current_dir;
uint8_t game_won;

uint16_t score;
uint16_t target_score;

uint16_t games;
uint16_t games_won;

void setup_buttons()
{
    // PC0 = buton de start/reset
    DDRC &= ~(1 << PC0);    
    PORTC |= (1 << PC0);

    // PC1 = buton stanga
    DDRC &= ~(1 << PC1);
    PORTC |= (1 << PC1);   

    // PC2 = buton sus
    DDRC &= ~(1 << PC2);
    PORTC |= (1 << PC2);   
    
    // PC3 = buton jos
    DDRC &= ~(1 << PC3);
    PORTC |= (1 << PC3);

    // PC4 = buton dreapta
    DDRC &= ~(1 << PC4);
    PORTC |= (1 << PC4);   

    // PD7 = led
    DDRD |= (1 << PD7);  
    // initial stins   
    PORTD &= ~(1 << PD7);


    // PD6 = buzzer
    DDRD |= (1 << PD6);  
    // initial fara nota   
    PORTD &= ~(1 << PD6);

    // buzzer inactiv
    OCR2B = 255;

    // activeaza semnal PWM pe pin
    TCCR2A |= (1 << COM2B1);

    // fast pwm
    TCCR2A |= (1 << WGM20) | (1 << WGM21);

    // prescaler 128
    TCCR2B |= (1 << CS22) | (1 << CS20); 
}

void check_buttons()
{
    // PC0 = buton de start/reset    
    if (!(PINC & (1 << PC0))) {
        if (prev_state == NONE) {
            // start
            prev_state = current_state;
            current_state = START_SCREEN;
            return;
        } else {
            // restart
            prev_state = END_SCREEN;
            current_state = START_SCREEN;
            return;
        }
    }

    // PC1 = buton stanga
    // pot face stanga doar daca nu merg la dreapta
    if (!(PINC & (1 << PC1)) && current_dir != RIGHT) {
        current_dir = LEFT;
        return;
    }
    
    // PC2 = buton sus
    // pot merge in sus doar daca nu merg in jos
    if (!(PINC & (1 << PC2)) && current_dir != DOWN) {
        current_dir = UP;        
        return;
    }
    
    // PC3 = buton jos
    // pot merge in jos doar daca nu merg in sus
    if (!(PINC & (1 << PC3)) && current_dir != UP) {
        current_dir = DOWN;    
        return;
    }
    
    // PC4 = buton dreapta
    // pot face dreapta doar daca nu merg la stanga
    if (!(PINC & (1 << PC4)) && current_dir != LEFT) {
        current_dir = RIGHT;
        return;
    } 
}

void start_sound_eat_food()
{
    // aprid led-ul
    PORTD |= (1 << PD7);

    // pornesc buzzer
    OCR2B = 128;
}

void stop_sound_eat_food()
{
    // opresc led
    PORTD &= ~(1 << PD7);

    // opresc buzzer
    OCR2B = 255;
}

void start_sound_game_won()
{
    // aprid led-ul
    PORTD |= (1 << PD7);

    // pornesc buzzer
    OCR2B = 192;
}

void stop_sound_game_won()
{
    // opresc buzzer
    OCR2B = 255;
}

void start_sound_game_lost()
{
    // pornesc buzzer
    OCR2B = 64;
}

void stop_sound_game_lost()
{
    // opresc buzzer
    OCR2B = 255;
}


void setup_lcd()
{
    ST7735R_Begin();
    ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, (uint8_t) 255, (uint8_t) 255, (uint8_t) 255);
}

void setup_food()
{
    // generez o pozitie random
    // daca pic pe sarpe, generez alta
    while (1) {
        uint8_t max_lin = ST7735R_HEIGHT - 10, max_col = ST7735R_WIDTH - 10;

        uint8_t lin = (rand() % max_lin) + 10;
        uint8_t col = (rand() % max_col) + 10;

        int on_snake = 0;
        for (uint8_t i = 0; i < snake_len; ++i) {
            if (snake[i].lin == lin && snake[i].col == col) {
                // am picat pe sarpe
                on_snake = 1;
                break;
            }
        }
        if (on_snake)
            // generez alta pozitie
            continue;

        // pozitie valida generata
        food.lin = lin;
        food.col = col;
        break;
    }


    // generez culoarea pentru mancare
    while (1) {
        food.r = (uint8_t) (rand() % 256);
        food.g = (uint8_t) (rand() % 256);
        food.b = (uint8_t) (rand() % 256);

        if (food.r + food.g + food.b <= 600 || food.r + food.g + food.b >= 200) 
            break;
    }

    regenerate_time = 10000 - 500 * games_won;
    if (regenerate_time < 5000)
        regenerate_time = 5000;
}

void setup_snake() {
    // initial sarpele are o lungime predefinita
    snake_len = SNAKE_START_SIZE;

    // si merge la dreapta
    current_dir = RIGHT;

    uint16_t offset = (SNAKE_START_SIZE - 1) * PIXEL_DIM;
    for (uint8_t i = 0; i < snake_len; ++i) {
        snake[i].lin = (ST7735R_HEIGHT / 2);
        snake[i].col = (ST7735R_WIDTH / 4) + offset;
        
        // sarpele este negru
        snake[i].r = (uint8_t) 0;
        snake[i].g = (uint8_t) 0;
        snake[i].b = (uint8_t) 0;

        offset -= PIXEL_DIM;
    }
}

void print_start_screen()
{
    char welcome_buffer[BUFFER_SIZE];
    sprintf(welcome_buffer, "Let's play SNAKE!");

    char level_buffer[BUFFER_SIZE];
    sprintf(level_buffer, "Level %2d", (games_won + 1));

    // print welcome message
    ST7735R_DrawText(
        10, 75,
        (const char*)welcome_buffer ,
        (uint8_t) 0, (uint8_t) 255, (uint8_t) 0,
        (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
    );

    // print level
    ST7735R_DrawText(
        45, 85,
        (const char*)level_buffer ,
        (uint8_t) 0, (uint8_t) 255, (uint8_t) 0,
        (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
    );
}

void print_end_screen()
{
    char game_status[BUFFER_SIZE];
    if (game_won) {
        sprintf(game_status, "You won!");
    } else {
        sprintf(game_status, "You lost!");
    }

    char game_replay[BUFFER_SIZE];
    sprintf(game_replay, "Play level %2d", (games_won + 1));

    // ecran negru
    ST7735R_FillRect(
        0, 0,
        ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1,
        (uint8_t) 0, (uint8_t) 0, (uint8_t) 0
    );

    // afisez mesajul won/lost
    ST7735R_DrawText(
        50, 75,
        (const char*) game_status,
        (uint8_t) 255, (uint8_t) 255, (uint8_t) 255,
        (uint8_t) 0, (uint8_t) 0, (uint8_t) 0
    );

    // afisez instructiunile
    ST7735R_DrawText(
        20, 85,
        (const char*) game_replay,
        (uint8_t) 255, (uint8_t) 0, (uint8_t) 0,
        (uint8_t) 0, (uint8_t) 0, (uint8_t) 0
    );
}

void print_game()
{
    char header_buff[BUFFER_SIZE];
    sprintf(header_buff, "Score:%2d/%2d G:%2d/%2d",
        score, target_score,
        games_won, games
    );

    char footer_buff[BUFFER_SIZE];
    sprintf(footer_buff, "%02d:%02d", regenerate_time / 1000, (regenerate_time % 1000)/100);

    // afisez header
    ST7735R_DrawText(
        2, 2,
        (const char*)header_buff,
        (uint8_t) 250, (uint8_t) 150, (uint8_t) 0,
        (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
    );
 
    // afisez footer
    ST7735R_DrawText(
        ST7735R_WIDTH / 2 - 10, ST7735R_HEIGHT - 10,
        (const char*)footer_buff,
        (uint8_t) 250, (uint8_t) 150, (uint8_t) 0,
        (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
    );

    // afisez capul (doar pixelul asta apare nou)
    ST7735R_FillRect(
        snake[0].col - BLOCK_OFFSET, snake[0].lin - BLOCK_OFFSET,
        snake[0].col - BLOCK_OFFSET + PIXEL_DIM, snake[0].lin - BLOCK_OFFSET + PIXEL_DIM,
        snake[0].r, snake[0].g, snake[0].b
    );

    // afisez mancarea
    ST7735R_FillRect(
        food.col - BLOCK_OFFSET, food.lin - BLOCK_OFFSET,
        food.col - BLOCK_OFFSET + PIXEL_DIM, food.lin - BLOCK_OFFSET + PIXEL_DIM,
        food.r, food.g, food.b
    );
}

pixel_t  get_new_head(pixel_t old_head) {
    pixel_t new_head = old_head;

    // mereu deplaseaza se face cu un pixel => PIXEL_DIM
    if (current_dir == UP) {
        new_head.lin -= PIXEL_DIM;
    } else if (current_dir == DOWN) {
        new_head.lin += PIXEL_DIM;
    } else if (current_dir == LEFT) {
        new_head.col -= PIXEL_DIM;
    } else if (current_dir == RIGHT) {
        new_head.col += PIXEL_DIM;
    }

    if (new_head.lin > ST7735R_HEIGHT) {
        // verific daca am depasit high, atunci "teleportez" capul pe linia 0
        new_head.lin = 0;
    } else if (new_head.lin < 0) {
        // verific daca am ajuns sub 0, atunci "teleportez" capul pe ultima linie
        new_head.lin = ST7735R_HEIGHT;
    }

    if (new_head.col > ST7735R_WIDTH) {
        // verific daca am depasit witdh, atunci "teleportez" capul pe coloana 0
        new_head.col = 0;
    } else if (new_head.col < 0) {
        // verific daca am ajusn sub 0, atunci "teleportez" capul pe ultima coloana
        new_head.col = ST7735R_WIDTH;
    }

    return new_head;
}

void move_snake() {
    pixel_t new_head = get_new_head(snake[0]);

    // verific cand sarpele se ciocneste de el
    for (uint8_t i = 0; i < snake_len; ++i) {
        if (snake[i].lin == new_head.lin && snake[i].col == new_head.col) {
            prev_state = current_state;
            current_state = END_SCREEN;

            start_sound_game_lost();
            _delay_ms(300);
            stop_sound_game_lost();

            return;
        }
    }

    // verific daca trebuie sa manance mancarea
    if (abs(new_head.lin - food.lin) < PIXEL_DIM && abs(new_head.col - food.col) < PIXEL_DIM) {
        // sarpele manca
        ++score;
        // si creste
        ++snake_len;

        // mancare consumata
        ST7735R_FillRect(
            food.col - BLOCK_OFFSET, food.lin - BLOCK_OFFSET,
            food.col - BLOCK_OFFSET + PIXEL_DIM, food.lin - BLOCK_OFFSET + PIXEL_DIM,
            (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
        );
        
        // regenerez mancare
        setup_food();

        start_sound_eat_food();
        _delay_ms(50);
        stop_sound_eat_food();
    }

    // jocul se termina daca
    //  - scorul nivelului a fost atins
    //      sau
    //  - sarpele a atins dimensiunea maxima
    if (score == target_score || snake_len > SNAKE_MAX_SIZE) {
        current_state = END_SCREEN;
        game_won = 1;

        ++games_won;
        target_score *= 2;

        start_sound_game_won();
        _delay_ms(500);
        stop_sound_game_won();
        return;
    }

    // sarpele se muta - se vor modifi doar 2 pixeli
    // 1. unde e coada acum va  fi liber
    // 2. capul va ajunge pe o celula noua

    // sterg coada de pe ecran (fac alb)
    ST7735R_FillRect(
        snake[snake_len - 1].col - BLOCK_OFFSET,
        snake[snake_len - 1].lin - BLOCK_OFFSET,
        snake[snake_len - 1].col - BLOCK_OFFSET + PIXEL_DIM,
        snake[snake_len - 1].lin - BLOCK_OFFSET + PIXEL_DIM,
        (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
    );

    // ceilalti pixeli raman desenati, insa au indici mai mari (se duc in spate)
    for (uint8_t i = snake_len - 1; i > 0; --i) {
        snake[i] = snake[i - 1];
    }

    // sarpele a inaintat o pozitie
    snake[0] = new_head;
}

void reset_game() {
    prev_state = START_SCREEN;
    current_state = PLAY_SCREEN;
    current_dir = RIGHT;
    
    game_won = 0;

    ++games;
    setup_snake();
    setup_food();

    score = 0;

    // fac ecranul alb
    ST7735R_FillRect(
        0, 0,
        ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1,
        (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
    );
}

void play_game()
{
    if (regenerate_time < 0) {
        // sterge mancarea
        ST7735R_FillRect(
            food.col - BLOCK_OFFSET, food.lin - BLOCK_OFFSET,
            food.col - BLOCK_OFFSET + PIXEL_DIM, food.lin - BLOCK_OFFSET + PIXEL_DIM,
            (uint8_t) 255, (uint8_t) 255, (uint8_t) 255
        );

        // regenereaza alta
        setup_food();
    }

    move_snake();
    print_game();
    _delay_ms(100);
    regenerate_time -= 100;
}

int play_snake()
{
    SPI_init();
    setup_buttons();
    setup_lcd();

    prev_state = NONE;
    current_state = START_SCREEN;    
    games = 0;
    games_won = 0;
    target_score = 2;

    while (1) {
        check_buttons();

        if (prev_state == NONE && current_state == START_SCREEN) {
            print_start_screen();
            _delay_ms(300);
            continue;
        }

        if (prev_state == END_SCREEN && current_state == START_SCREEN) {
            reset_game();
            _delay_ms(300);
            continue;
        }

        if (current_state == PLAY_SCREEN) {
            play_game();
            continue;
        }

        if (current_state == END_SCREEN) {
            print_end_screen();
            _delay_ms(300);
            continue;
        }
    }
}

int main()
{
    return play_snake();
}
